import json
from typing import Optional, Dict


class SingleDetectionResult(dict):
    """A span of text that has been detected as a named entity.

    Attributes
    ----------
    start : int
        The start index of the entity in the original text
    end : int
        The end index of the entity in the original text. The end index is exclusive.
    label : str
        The label of the entity
    text : str
        The substring of the original text that was detected as an entity.
    score : float
        The confidence score of the detection
    json_path : Optional[str]
        The JSON path of the entity in the original JSON document. This is only
        present if the input text was a JSON document.
    """

    def __init__(
        self,
        start: int,
        end: int,
        label: str,
        text: str,
        score: float,
        json_path: Optional[str] = None,
    ):
        self.start = start
        self.end = end
        self.label = label
        self.text = text
        self.score = score
        self.jsonPath = json_path
        if json_path is None:
            dict.__init__(
                self, start=start, end=end, label=label, text=text, score=score
            )
        else:
            dict.__init__(
                self,
                start=start,
                end=end,
                label=label,
                text=text,
                score=score,
                json_path=json_path,
            )

    def describe(self) -> str:
        return json.dumps(self.to_dict(), indent=2)

    def to_dict(self) -> Dict:
        out = {
            "start": self.start,
            "end": self.end,
            "label": self.label,
            "text": self.text,
            "score": self.score,
        }
        if self.jsonPath is not None:
            out["jsonPath"] = self.jsonPath
        return out



from enum import Enum


class FileTypeEnum(str, Enum):
    csv = ("Csv",)
    raw = ("Raw",)
    pdf = ("Pdf",)
    docX = ("DocX",)
    png = ("Png",)
    jpg = ("Jpg",)
    tif = ("Tif",)
    xlsx = ("Xlsx",)



from tonic_textual.classes.enums.file_source import FileSource


class PipelineDatabricksCredential(dict):
    """Class to represent Databricks credentials used to create a Databricks pipeline.

    Parameters
    ----------
    url: str
        The URL to your Databricks instance.

    access_token: str
        The access token. The token must have access to the Unity Catalog that you plan to read data from and write data to.
    """

    def __init__(self, url: str, access_token: str):
        self.file_source = FileSource.databricks
        self.url = url
        self.access_token = access_token

        dict.__init__(self, url=url, accessToken=access_token)

    def to_dict(self):
        return {"url": self.url, "accessToken": self.access_token}



import io
import json
import os
import requests

from time import sleep
from typing import List, Optional, Union, Dict
from urllib.parse import urlencode

from tonic_textual.classes.common_api_responses.label_custom_list import LabelCustomList
from tonic_textual.classes.common_api_responses.replacement import Replacement
from tonic_textual.classes.common_api_responses.single_detection_result import (
    SingleDetectionResult,
)
from tonic_textual.classes.httpclient import HttpClient
from tonic_textual.classes.record_api_request_options import RecordApiRequestOptions
from tonic_textual.classes.redact_api_responses.bulk_redaction_response import BulkRedactionResponse
from tonic_textual.classes.redact_api_responses.redaction_response import (
    RedactionResponse,
)
from tonic_textual.enums.pii_state import PiiState
from tonic_textual.services.dataset import DatasetService
from tonic_textual.services.datasetfile import DatasetFileService
from tonic_textual.classes.dataset import Dataset
from tonic_textual.classes.datasetfile import DatasetFile
from tonic_textual.classes.tonic_exception import (
    BadArgumentsException,
    DatasetNameAlreadyExists,
    InvalidJsonForRedactionRequest,
    FileNotReadyForDownload,
)

from tonic_textual.generator_utils import validate_generator_options


class TextualNer:
    """Wrapper class to invoke the Tonic Textual API

    Parameters
    ----------
    base_url : str
        The URL to your Tonic Textual instance. Do not include trailing backslashes. The default value is https://textual.tonic.ai.
    api_key : str
        Optional. Your API token. Instead of providing the API token
        here, we recommended that you set the API key in your environment as the
        value of TONIC_TEXTUAL_API_KEY.
    verify: bool
        Whether to verify SSL certification. By default, this is enabled.
    Examples
    --------
    >>> from tonic_textual.redact_api import TextualNer
    >>> textual = TonicTextual("https://textual.tonic.ai")
    """

    default_record_options = RecordApiRequestOptions(False, 0, [])

    def __init__(
        self, base_url: str = "https://textual.tonic.ai", api_key: Optional[str] = None, verify: bool = True
    ):
        if api_key is None:
            api_key = os.environ.get("TONIC_TEXTUAL_API_KEY")
            if api_key is None:
                raise Exception(
                    "No API key provided. Either provide an API key, or set the API "
                    "key as the value of the TONIC_TEXTUAL_API_KEY environment "
                    "variable."
                )
            

        self.api_key = api_key
        self.client = HttpClient(base_url, self.api_key, verify)
        self.dataset_service = DatasetService(self.client)
        self.datasetfile_service = DatasetFileService(self.client)
        self.verify = verify

    def create_dataset(self, dataset_name: str):
        """Creates a dataset. A dataset is a collection of 1 or more files for Tonic
        Textual to scan and redact.

        Parameters
        -----
        dataset_name : str
            The name of the dataset. Dataset names must be unique.


        Returns
        -------
        Dataset
            The newly created dataset.


        Raises
        ------

        DatasetNameAlreadyExists
            Raised if a dataset with the same name already exists.

        """

        try:
            self.client.http_post("/api/dataset", data={"name": dataset_name})
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 409:
                raise DatasetNameAlreadyExists(e)

        return self.get_dataset(dataset_name)

    def delete_dataset(self, dataset_name: str):
        """Deletes dataset by name.

        Parameters
        -----
        dataset_name : str
            The name of the dataset to delete.
        """

        params = {"datasetName": dataset_name}
        self.client.http_delete(
            "/api/dataset/delete_dataset_by_name?" + urlencode(params)
        )

    def get_dataset(self, dataset_name: str) -> Dataset:
        """Gets the dataset for the specified dataset name.

        Parameters
        ----------
        dataset_name : str
            The name of the dataset.

        Returns
        -------
        Dataset

        Examples
        --------
        >>> dataset = tonic.get_dataset("llama_2_chatbot_finetune_v5")
        """

        return self.dataset_service.get_dataset(dataset_name)

    def get_files(self, dataset_id: str) -> List[DatasetFile]:
        """
        Gets all of the files in the dataset.

        Returns
        ------
        List[DatasetFile]
            A list of all of the files in the dataset.
        """

        return self.datasetfile_service.get_files(dataset_id)

    def unredact_bulk(
        self, redacted_strings: List[str], random_seed: Optional[int] = None
    ) -> List[str]:
        """Removes redaction from a list of strings. Returns the strings with the
        original values.

        Parameters
        ----------
        redacted_strings : List[str]
            The list of redacted strings from which to remove the redaction.

        random_seed: Optional[int] = None
            Ann optional value to use to override Textual's default random number
            seeding. Can be used to ensure that different API calls use the same or
            different random seeds.

        Returns
        -------
        List[str]
            The list of strings with the redaction removed.
        """

        if random_seed is not None:
            additional_headers = {"textual-random-seed": str(random_seed)}
        else:
            additional_headers = {}

        response = self.client.http_post(
            "/api/unredact",
            data=redacted_strings,
            additional_headers=additional_headers,
        )
        return response

    def unredact(self, redacted_string: str, random_seed: Optional[int] = None) -> str:
        """Removes the redaction from a provided string. Returns the string with the
        original values.

        Parameters
        ----------
        redacted_string : str
            The redacted string from which to remove the redaction.

        random_seed: Optional[int] = None
            An optional value to use to override Textual's default random number
            seeding. Can be used to ensure that different API calls use the same or
            different random seeds.

        Returns
        -------
        str
            The string with the redaction removed.
        """

        if random_seed is not None:
            additional_headers = {"textual-random-seed": str(random_seed)}
        else:
            additional_headers = {}

        response = self.client.http_post(
            "/api/unredact",
            data=[redacted_string],
            additional_headers=additional_headers,
        )

        return response

    def redact(
        self,
        string: str,
        generator_config: Dict[str, PiiState] = dict(),
        generator_default: PiiState = PiiState.Redaction,
        random_seed: Optional[int] = None,
        label_block_lists: Optional[Dict[str, List[str]]] = None,
        label_allow_lists: Optional[Dict[str, List[str]]] = None,
        record_options: RecordApiRequestOptions = default_record_options
    ) -> RedactionResponse:
        """Redacts a string. Depending on the configured handling for each sensitive
        data type, values are either redacted, synthesized, or ignored.

        Parameters
        ----------
        string : str
            The string to redact.

        generator_config: Dict[str, PiiState]
            A dictionary of sensitive data entities. For each entity, indicates whether
            to redact, synthesize, or ignore it.
            Values must be one of "Redaction", "Synthesis", or "Off".

        generator_default: PiiState = PiiState.Redaction
            The default redaction used for types that are not specified in generator_config.
            Value must be one of "Redaction", "Synthesis", or "Off".

        random_seed: Optional[int] = None
            An optional value to use to override Textual's default random number
            seeding. Can be used to ensure that different API calls use the same or
            different random seeds.

        label_block_lists: Optional[Dict[str, List[str]]]
            A dictionary of (entity type, ignored values). When a value for an entity type matches a listed regular expression,
            the value is ignored and is not redacted or synthesized.

        label_allow_lists: Optional[Dict[str, List[str]]]
            A dictionary of (entity type, additional values). When a piece of text matches a listed regular expression,
            the text is marked as the entity type and is included in the redaction or synthesis.

        record_options: RecordApiRequestOptions
            A value to record the API request and results for analysis in the Textual application. The default value is to not record the API request.
            
        Returns
        -------
        RedactionResponse
            The redacted string along with ancillary information.

        Examples
        --------
            >>> textual.redact(
            >>>     "John Smith is a person",
            >>>     # only redacts NAME_GIVEN
            >>>     generator_config={"NAME_GIVEN": "Redaction"},
            >>>     generator_default="Off",
            >>>     # Occurrences of "There" are treated as NAME_GIVEN entities
            >>>     label_allow_lists={"NAME_GIVEN": ["There"]},
            >>>     # Text matching the regex ` ([a-z]{2}) ` is not treated as an occurrence of NAME_FAMILY
            >>>     label_block_lists={"NAME_FAMILY": [" ([a-z]{2}) "]},
            >>> )


        """

        validate_generator_options(generator_default, generator_config)
        payload = {
            "text": string,
            "generatorDefault": generator_default,
            "generatorConfig": generator_config,
        }

        if label_block_lists is not None:
            payload["labelBlockLists"] = {
                k: LabelCustomList(regexes=v).to_dict()
                for k, v in label_block_lists.items()
            }
        if label_allow_lists is not None:
            payload["labelAllowLists"] = {
                k: LabelCustomList(regexes=v).to_dict()
                for k, v in label_allow_lists.items()
            }

        if record_options.record:
            if record_options.retention_time_in_hours <= 0 or record_options.retention_time_in_hours>720:
                raise BadArgumentsException("The retention time must be set between 1 and 720 hours")
            
            record_payload = { "retentionTimeInHours": record_options.retention_time_in_hours, "tags": record_options.tags, "record": True}
            payload["recordApiRequestOptions"] = record_payload
        else:
            payload["recordApiRequestOptions"] = None
        
        return self.send_redact_request("/api/redact", payload, random_seed)
    
    def redact_bulk(
        self,
        strings: List[str],
        generator_config: Dict[str, PiiState] = dict(),
        generator_default: PiiState = PiiState.Redaction,
        random_seed: Optional[int] = None,
        label_block_lists: Optional[Dict[str, List[str]]] = None,
        label_allow_lists: Optional[Dict[str, List[str]]] = None
    ) -> BulkRedactionResponse:
        """Redacts a string. Depending on the configured handling for each sensitive
        data type, values are either redacted, synthesized, or ignored.

        Parameters
        ----------
        strings : List[str]
            The array of strings to redact.

        generator_config: Dict[str, PiiState]
            A dictionary of sensitive data entities. For each entity, indicates whether
            to redact, synthesize, or ignore it.
            Values must be one of "Redaction", "Synthesis", or "Off".

        generator_default: PiiState = PiiState.Redaction
            The default redaction used for all types that are not specified in generator_config.
            Value must be one of "Redaction", "Synthesis", or "Off".

        random_seed: Optional[int] = None
            An optional value to use to override Textual's default random number
            seeding. Can be used to ensure that different API calls use the same or
            different random seeds.

        label_block_lists: Optional[Dict[str, List[str]]]
            A dictionary of (entity type, ignored values). When a value for an entity type matches a listed regular expression,
            the value is ignored and is not redacted or synthesized.

        label_allow_lists: Optional[Dict[str, List[str]]]
            A dictionary of (entity type, additional values). When a piece of text matches a listed regular expression,
            the text is marked as the entity type and is included in the redaction or synthesis.


        Returns
        -------
        RedactionResponse
            The redacted string along with ancillary information.

        Examples
        --------
            >>> textual.redact_bulk(
            >>>     ["John Smith is a person", "I live in Atlanta"],
            >>>     # only redacts NAME_GIVEN
            >>>     generator_config={"NAME_GIVEN": "Redaction"},
            >>>     generator_default="Off",
            >>>     # Occurrences of "There" are treated as NAME_GIVEN entities
            >>>     label_allow_lists={"NAME_GIVEN": ["There"]},
            >>>     # Text matching the regex ` ([a-z]{2}) ` is not treated as an occurrence of NAME_FAMILY
            >>>     label_block_lists={"NAME_FAMILY": [" ([a-z]{2}) "]},
            >>> )


        """

        validate_generator_options(generator_default, generator_config)
        payload = {
            "bulkText": strings,
            "generatorDefault": generator_default,
            "generatorConfig": generator_config,
        }

        if label_block_lists is not None:
            payload["labelBlockLists"] = {
                k: LabelCustomList(regexes=v).to_dict()
                for k, v in label_block_lists.items()
            }
        if label_allow_lists is not None:
            payload["labelAllowLists"] = {
                k: LabelCustomList(regexes=v).to_dict()
                for k, v in label_allow_lists.items()
            }

        return self.send_redact_bulk_request("/api/redact/bulk", payload, random_seed)

    def llm_synthesis(
        self,
        string: str,
        generator_config: Dict[str, PiiState] = dict(),
        generator_default: PiiState = PiiState.Redaction,
    ) -> RedactionResponse:
        """Deidentifies a string. Redacting sensitive data and replaces those values
        with values generated by an LLM.

        Parameters
        ----------
        string: str
                The string to redact.

        generator_config: Dict[str, PiiState]
                A dictionary of sensitive data entities. For each entity, indicates
                whether to redact, synthesize, or ignore it.

        generator_default: PiiState = PiiState.Redaction
            The default redaction used for all types that are not specified in generator_config.

        Returns
        -------
        RedactionResponse
            The redacted string, along with ancillary information about the detected entities.
        """
        validate_generator_options(generator_default, generator_config)
        endpoint = "/api/synthesis"
        response = self.client.http_post(
            endpoint,
            data={
                "text": string,
                "generatorDefault": generator_default,
                "generatorConfig": generator_config,
            },
        )

        de_id_results = [
            SingleDetectionResult(
                x["start"], x["end"], x["label"], x["text"], x["score"]
            )
            for x in list(response["deIdentifyResults"])
        ]

        return RedactionResponse(
            response["originalText"],
            response["redactedText"],
            response["usage"],
            de_id_results,
        )

    def redact_json(
        self,
        json_data: Union[str, dict],
        generator_config: Dict[str, PiiState] = dict(),
        generator_default: PiiState = PiiState.Redaction,
        random_seed: Optional[int] = None,
        label_block_lists: Optional[Dict[str, List[str]]] = None,
        label_allow_lists: Optional[Dict[str, List[str]]] = None,
        jsonpath_allow_lists: Optional[Dict[str, List[str]]] = None,
    ) -> RedactionResponse:
        """Redacts the values in a JSON blob. Depending on the configured handling for
        each sensitive data type, values are either redacted, synthesized, or
        ignored.

        Parameters
        ----------
        json_string : Union[str, dict]
            The JSON for which to redact values. This can be either a JSON string
            or a Python dictionary.

        generator_config: Dict[str, PiiState]
            A dictionary of sensitive data entities. For each entity, indicates whether
            to redact, synthesize, or ignore it.

        generator_default: PiiState = PiiState.Redaction
            The default redaction to use for all types that are not specified in generator_config.

        random_seed: Optional[int] = None
            An optional value to use to override Textual's default random number
            seeding. Can be used to ensure that different API calls use the same or
            different random seeds.

        label_block_lists: Optional[Dict[str, List[str]]]
            A dictionary of (entity type, ignored values). When an value for the entity type matches a listed regular expression,
            the value is ignored and is not redacted or synthesized.

        label_allow_lists: Optional[Dict[str, List[str]]]
            A dictionary of (entity type, additional values). When a piece of text matches a listed regular expression,
            the text is marked as the entity type and is included in the redaction or synthesis.

        jsonpath_allow_lists: Optional[Dict[str, List[str]]]
            A dictionary of (entity type, path expression). When an element in the JSON document matches the JSON path expression, the entire text value is treated as the specified entity type.
            Only supported for path expressions that point to JSON primitive values. This setting overrides any results found by the NER model or in label allow and block lists.
            If multiple path expressions point to the same JSON node, but specify different entity types, then the value is redacted as one of those types. However, the chosen type is selected at random - it could use any of the types.

        Returns
        -------
        RedactionResponse
            The redacted string along with ancillary information.
        """
        validate_generator_options(generator_default, generator_config)

        if isinstance(json_data, str):
            json_text = json_data
        elif isinstance(json_data, dict):
            json_text = json.dumps(json_data)
        else:
            raise Exception(
                "redact_json must receive either a JSON blob as a string or dict(). "
                f"You passed in type {type(json_data)} which is not supported"
            )
        payload = {
            "jsonText": json_text,
            "generatorDefault": generator_default,
            "generatorConfig": generator_config,
        }
        if label_block_lists is not None:
            payload["labelBlockLists"] = {
                k: LabelCustomList(regexes=v).to_dict()
                for k, v in label_block_lists.items()
            }
        if label_allow_lists is not None:
            payload["labelAllowLists"] = {
                k: LabelCustomList(regexes=v).to_dict()
                for k, v in label_allow_lists.items()
            }
        if jsonpath_allow_lists is not None:
            payload["jsonPathAllowLists"] = jsonpath_allow_lists
        return self.send_redact_request("/api/redact/json", payload, random_seed)

    def redact_xml(
        self,
        xml_data: str,
        generator_config: Dict[str, PiiState] = dict(),
        generator_default: PiiState = PiiState.Redaction,
        random_seed: Optional[int] = None,
        label_block_lists: Optional[Dict[str, List[str]]] = None,
        label_allow_lists: Optional[Dict[str, List[str]]] = None,
    ) -> RedactionResponse:
        """Redacts the values in an XML blob. Depending on the configured handling for
        each entity type, values are either redacted, synthesized, or
        ignored.

        Parameters
        ----------
        xml_data : str
            The XML for which to redact values.

        generator_config: Dict[str, PiiState]
            A dictionary of entity types. For each entity type, indicates
            whether to redact, synthesize, or ignore the detected values.

        generator_default: PiiState = PiiState.Redaction
            The default redaction used for any entity type that is not included in generator_config.

        random_seed: Optional[int] = None
            An optional value to use to override Textual's default random number
            seeding. Can be used to ensure that different API calls use the same or
            different random seeds.

        label_block_lists: Optional[Dict[str, List[str]]]
            A dictionary of (entity type, ignored values). When an value for the entity type matches a listed regular expression,
            the value is ignored and is not redacted or synthesized.

        label_allow_lists: Optional[Dict[str, List[str]]]
            A dictionary of (entity type, additional values). When a piece of text matches a listed regular expression,
            the text is marked as the entity type and is included in the redaction or synthesis.

        Returns
        -------
        RedactionResponse
            The redacted string plus additional information.
        """
        validate_generator_options(generator_default, generator_config)

        payload = {
            "xmlText": xml_data,
            "generatorDefault": generator_default,
            "generatorConfig": generator_config,
        }

        if label_block_lists is not None:
            payload["labelBlockLists"] = {
                k: LabelCustomList(regexes=v).to_dict()
                for k, v in label_block_lists.items()
            }
        if label_allow_lists is not None:
            payload["labelAllowLists"] = {
                k: LabelCustomList(regexes=v).to_dict()
                for k, v in label_allow_lists.items()
            }

        return self.send_redact_request("/api/redact/xml", payload, random_seed)

    def redact_html(
        self,
        html_data: str,
        generator_config: Dict[str, PiiState] = dict(),
        generator_default: PiiState = PiiState.Redaction,
        random_seed: Optional[int] = None,
        label_block_lists: Optional[Dict[str, List[str]]] = None,
        label_allow_lists: Optional[Dict[str, List[str]]] = None,
    ) -> RedactionResponse:
        """Redacts the values in an HTML blob. Depending on the configured handling for
        each entity type, values are either redacted, synthesized, or
        ignored.

        Parameters
        ----------
        html_data : str
            The HTML for which to redact values.

        generator_config: Dict[str, PiiState]
            A dictionary of entity types. For each entity type, indicates
            whether to redact, synthesize, or ignore the detected values.

        generator_default: PiiState = PiiState.Redaction
            The default redaction used for any entity type that is not included in generator_config.

        random_seed: Optional[int] = None
            An optional value to use to override Textual's default random number
            seeding. Can be used to ensure that different API calls use the same or
            different random seeds.

        label_block_lists: Optional[Dict[str, List[str]]]
            A dictionary of (entity type, ignored values). The ignored values are regular expressions. When a value for the entity type matches a listed regular expression,
            the value is ignored and is not redacted or synthesized.

        label_allow_lists: Optional[Dict[str, List[str]]]
            A dictionary of (entity type, additional values). The additional values are regular expressions. When a piece of text matches a listed regular expression,
            the text is marked as the entity type and is included in the redaction or synthesis.

        Returns
        -------
        RedactionResponse
            The redacted string plus additional information.
        """
        validate_generator_options(generator_default, generator_config)

        payload = {
            "htmlText": html_data,
            "generatorDefault": generator_default,
            "generatorConfig": generator_config,
        }

        if label_block_lists is not None:
            payload["labelBlockLists"] = {
                k: LabelCustomList(regexes=v).to_dict()
                for k, v in label_block_lists.items()
            }
        if label_allow_lists is not None:
            payload["labelAllowLists"] = {
                k: LabelCustomList(regexes=v).to_dict()
                for k, v in label_allow_lists.items()
            }

        return self.send_redact_request("/api/redact/html", payload, random_seed)

    def send_redact_request(
        self,
        endpoint: str,
        payload: Dict,
        random_seed: Optional[int] = None,
    ) -> RedactionResponse:
        """Helper function to send redact requests, handle responses, and catch errors."""

        if random_seed is not None:
            additional_headers = {"textual-random-seed": str(random_seed)}
        else:
            additional_headers = {}

        try:
            response = self.client.http_post(
                endpoint, data=payload, additional_headers=additional_headers
            )
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 400:
                raise InvalidJsonForRedactionRequest(e.response.text)
            raise e

        de_id_results = [
            Replacement(
                start=x["start"],
                end=x["end"],
                new_start=x.get("newStart"),
                new_end=x.get("newEnd"),
                label=x["label"],
                text=x["text"],
                new_text=x.get("newText"),
                score=x["score"],
                language=x.get("language"),
                example_redaction=x.get("exampleRedaction"),
                json_path=x.get("jsonPath"),
                xml_path=x.get("xmlPath"),
            )
            for x in response["deIdentifyResults"]
        ]

        return RedactionResponse(
            response["originalText"],
            response["redactedText"],
            response["usage"],
            de_id_results,
        )

    def send_redact_bulk_request(
        self,
        endpoint: str,
        payload: Dict,
        random_seed: Optional[int] = None,
    ) -> BulkRedactionResponse:
        """Helper function to send redact requests, handle responses, and catch errors."""

        if random_seed is not None:
            additional_headers = {"textual-random-seed": str(random_seed)}
        else:
            additional_headers = {}

        try:
            response = self.client.http_post(
                endpoint, data=payload, additional_headers=additional_headers
            )
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 400:
                raise InvalidJsonForRedactionRequest(e.response.text)
            raise e
        
        de_id_results = [[] for i in range(len(response["bulkText"]))]
        for x in response["deIdentifyResults"]:
            de_id_results[x["idx"]].append(
                Replacement(
                    start=x["start"],
                    end=x["end"],
                    new_start=x.get("newStart"),
                    new_end=x.get("newEnd"),
                    label=x["label"],
                    text=x["text"],
                    new_text=x.get("newText"),
                    score=x["score"],
                    language=x.get("language"),
                    example_redaction=x.get("exampleRedaction"),
                )
            )

        return BulkRedactionResponse(
            response["bulkText"],
            response["bulkRedactedText"],
            response["usage"],
            de_id_results,
        )

    def start_file_redaction(self, file: io.IOBase, file_name: str) -> str:
        """
        Redact a provided file

        Parameters
        --------
        file: io.IOBase
            The opened file, available for reading, to upload and redact.
        file_name: str
            The name of the file.

        Returns
        -------
        str
           The job identifier, which can be used to download the redacted file when it is ready.

        """

        files = {
            "document": (
                None,
                json.dumps({"fileName": file_name, "csvConfig": {}, "datasetId": ""}),
                "application/json",
            ),
            "file": file,
        }

        response = self.client.http_post("/api/unattachedfile/upload", files=files)

        return response["jobId"]

    def download_redacted_file(
        self,
        job_id: str,
        generator_config: Dict[str, PiiState] = dict(),
        generator_default: PiiState = PiiState.Redaction,
        random_seed: Optional[int] = None,
        label_block_lists: Optional[Dict[str, List[str]]] = None,
        num_retries: int = 6,
        wait_between_retries: int = 10,
    ) -> bytes:
        """
        Download a redacted file

        Parameters
        --------
        job_id: str
            The identifier of the redaction job.

        generator_config: Dict[str, PiiState]
            A dictionary of sensitive data entities. For each entity, indicates whether
            to redact, synthesize, or ignore it.

        generator_default: PiiState = PiiState.Redaction
            The default redaction used for all types that are not specified in generator_config.

        random_seed: Optional[int] = None
            An optional value to use to override Textual's default random number
            seeding. Can be used to ensure that different API calls use the same or
            different random seeds.

        label_block_lists: Optional[Dict[str, List[str]]]
            A dictionary of (entity type, ignored values). When a value for the entity type matches a listed regular expression,
            the value is ignored and is not redacted or synthesized.

        num_retries: int = 6
            An optional value to specify the number of times to attempt to download the
            file. If a file is not yet ready for download, Textual pauses 10-second
            pause before retrying. (The default value is 6)

        wait_between_retries: int = 10
            The number of seconds to wait between retry attempts. (The default value is 10)

        Returns
        -------
        bytes
            The redacted file as a byte array.
        """

        validate_generator_options(generator_default, generator_config)
        retries = 1
        while retries <= num_retries:
            try:
                if random_seed is not None:
                    additional_headers = {"textual-random-seed": str(random_seed)}
                else:
                    additional_headers = {}
                data = {
                    "generatorDefault": generator_default,
                    "generatorConfig": generator_config,
                }
                if label_block_lists is not None:
                    data["labelBlockLists"] = {
                        k: LabelCustomList(regexes=v).to_dict()
                        for k, v in label_block_lists.items()
                    }
                return self.client.http_post_download_file(
                    f"/api/unattachedfile/{job_id}/download",
                    data=data,
                    additional_headers=additional_headers,
                )

            except FileNotReadyForDownload:
                retries = retries + 1
                if retries <= num_retries:
                    sleep(wait_between_retries)

        retryWord = "retry" if num_retries == 1 else "retries"
        raise FileNotReadyForDownload(
            f"After {num_retries} {retryWord}, the file is not yet ready to download. "
            "This is likely due to a high service load. Try again later."
        )


class TonicTextual(TextualNer):
    pass



from tonic_textual.classes.dataset import Dataset
from urllib.parse import urlencode
import requests


class DatasetService:
    def __init__(self, client):
        self.client = client

    def get_dataset(self, dataset_name):
        with requests.Session() as session:
            params = {"datasetName": dataset_name}
            dataset = self.client.http_get(
                "/api/dataset/get_dataset_by_name?" + urlencode(params), session=session
            )
            return Dataset(
                self.client,
                dataset["id"],
                dataset["name"],
                dataset["files"],
                dataset["generatorSetup"],
                dataset["labelBlockLists"],
                dataset["labelAllowLists"],
            )



from typing import List

from tonic_textual.classes.file_content.base_document import BaseDocument
from tonic_textual.classes.file_content.content import Content
from tonic_textual.classes.httpclient import HttpClient
from tonic_textual.classes.table import Table


class XlsxDocument(BaseDocument):
    def __init__(self, client: HttpClient, json_def):
        super().__init__(client, json_def)
        self.tables: List[Table] = [
            Table(
                [[Content(client, f) for f in row] for row in table["data"]],
                table["tableName"],
                table["header"],
            )
            for table in json_def["tables"]
        ]

    def get_cell_content(self, sheet_name: str, i: int, j: int) -> Content:
        return self.content[sheet_name][i][j]



import requests
from time import sleep
from typing import Optional, Dict

from tonic_textual.classes.common_api_responses.label_custom_list import LabelCustomList
from tonic_textual.classes.enums.file_redaction_policies import docx_image_policy, docx_comment_policy, \
    pdf_signature_policy
from tonic_textual.classes.httpclient import HttpClient
from tonic_textual.classes.tonic_exception import FileNotReadyForDownload


class DatasetFile:
    """
    Class to store the metadata for a dataset file.

    Parameters
    ----------
    id : str
        The identifier of the dataset file.

    name: str
        The file name of the dataset file.

    num_rows : long
        The number of rows in the dataset file.

    num_columns: int
        The number of columns in the dataset file.

    processing_status: string
        The status of the dataset file in the processing pipeline. Possible values are
        'Completed', 'Failed', 'Cancelled', 'Running', and 'Queued'.

    processing_error: string
        If the dataset file processing failed, a description of the issue that caused
        the failure.

    label_allow_lists: Dict[str, LabelCustomList]
        A dictionary of custom entity detection regular expressions for the dataset file. Each key is an entity type to detect,
        and each values is a LabelCustomList object, whose regular expressions should be recognized as the specified entity type.
    """

    def __init__(
        self,
        client: HttpClient,
        id: str,
        dataset_id: str,
        name: str,
        num_rows: Optional[int],
        num_columns: int,
        processing_status: str,
        processing_error: Optional[str],
        label_allow_lists: Optional[Dict[str, LabelCustomList]] = None,
        docx_image_policy_name: Optional[docx_image_policy] = docx_image_policy.redact,
        docx_comment_policy_name: Optional[docx_comment_policy] = docx_comment_policy.remove,
        pdf_signature_policy_name: Optional[pdf_signature_policy] = pdf_signature_policy.redact
    ):
        self.client = client
        self.id = id
        self.dataset_id = dataset_id
        self.name = name
        self.num_rows = num_rows
        self.num_columns = num_columns
        self.processing_status = processing_status
        self.processing_error = processing_error
        self.label_allow_lists = label_allow_lists
        self.docx_image_policy: docx_image_policy_name
        self.docx_comment_policy:  docx_comment_policy_name
        self.pdf_signature_policy: pdf_signature_policy_name

    def describe(self) -> str:
        """Returns the dataset file metadata as string. Includes the identifier, file
        name, number of rows, and number of columns."""
        description = f"File: {self.name} [{self.id}]\n"
        description += f"Number of rows: {self.num_rows}\n"
        description += f"Number of columns: {self.num_columns}\n"
        description += f"Status: {self.processing_status}\n"
        if self.processing_status != "" and self.processing_error is not None:
            description += f"Error: {self.processing_error}\n"
        return description

    def download(
        self,
        random_seed: Optional[int] = None,
        num_retries: int = 6,
        wait_between_retries: int = 10,
    ) -> bytes:
        """
        Download a redacted file

        Parameters
        --------
        random_seed: Optional[int] = None
            An optional value to use to override Textual's default random number
            seeding. Can be used to ensure that different API calls use the same or
            different random seeds.

        num_retries: int = 6
            An optional value to specify the number of times to attempt to download the
            file. If a file is not yet ready for download, there is a 10-second
            pause before retrying. (The default value is 6)

        wait_between_retries: int = 10
            The number of seconds to wait between retry attempts.

        Returns
        -------
        bytes
            The redacted file as a byte array.
        """
        retries = 1
        while retries <= num_retries:
            try:
                if random_seed is not None:
                    additional_headers = {"textual-random-seed": str(random_seed)}
                else:
                    additional_headers = {}
                with requests.Session() as session:
                    return self.client.http_get_file(
                        f"/api/dataset/{self.dataset_id}/files/{self.id}/download",
                        additional_headers=additional_headers,
                        session=session,
                    )

            except FileNotReadyForDownload:
                retries = retries + 1
                if retries <= num_retries:
                    sleep(wait_between_retries)

        retryWord = "retry" if num_retries == 1 else "retries"
        raise FileNotReadyForDownload(
            f"After {num_retries} {retryWord}, the file is not yet ready to download. "
            "This is likely due to a high service load. Try again later."
        )



import io
import json
import os
from typing import List, Optional, Union
from warnings import warn
import requests
from requests import RequestException

from tonic_textual.classes.azure_pipeline import AzurePipeline
from tonic_textual.classes.enums.aws_credentials_source import AwsCredentialsSource
from tonic_textual.classes.enums.file_source import FileSource
from tonic_textual.classes.enums.object_storage_type import ObjectStorageType
from tonic_textual.classes.httpclient import HttpClient
from tonic_textual.classes.local_pipeline import LocalPipeline
from tonic_textual.classes.parse_api_responses.file_parse_result import FileParseResult
from tonic_textual.classes.pipeline_aws_credential import PipelineAwsCredential
from tonic_textual.classes.pipeline import Pipeline
from tonic_textual.classes.pipeline_azure_credential import PipelineAzureCredential
from tonic_textual.classes.pipeline_databricks_credential import (
    PipelineDatabricksCredential,
)
from tonic_textual.classes.s3_pipeline import S3Pipeline
from tonic_textual.classes.tonic_exception import (
    BadArgumentsException,
    PipelineDeleteError,
    PipelineCreateError,
)


class TextualParse:
    """Wrapper class for invoking Tonic Textual API

    Parameters
    ----------
    base_url : Optional[str]
        The URL to your Tonic Textual instance. Do not include trailing backslashes. The default value is https://textual.tonic.ai.
    api_key : Optional[str]
        Optional. Your API token. Instead of providing the API token
        here, we recommended that you set the API key in your environment as the
        value of TEXTUAL_API_KEY.
    verify: bool
        Whether to verify SSL certification verification. By default, this is enabled.
    Examples
    --------
    >>> from tonic_textual.parse_api import TextualParse
    >>> textual = TonicTextualParse("https://textual.tonic.ai")
    """

    def __init__(
        self, base_url: str = "https://textual.tonic.ai", api_key: Optional[str] = None, verify: bool = True
    ):
        if api_key is None:
            api_key = os.environ.get("TONIC_TEXTUAL_API_KEY")
            if api_key is None:
                raise Exception(
                    "No API key provided. Either provide an API key, or set the API "
                    "key as the value of the TEXTUAL_API_KEY environment "
                    "variable."
                )

        self.api_key = api_key
        self.client = HttpClient(base_url, self.api_key, verify)
        self.verify = verify

    def get_pipelines(self) -> List[Pipeline]:
        """Get the pipelines for the Tonic Textual instance.

        Returns
        -------
        List[Pipeline]
            A list of pipeline objects, ordered by their creation timestamp.
        Examples
        --------
        >>> latest_pipeline = textual.get_pipelines()[-1]
        """

        with requests.Session() as session:
            response = self.client.http_get("/api/parsejobconfig", session=session)
            pipelines: List[Pipeline] = []
            for x in response:
                pipelines.append(Pipeline(x["name"], x["id"], self.client))
            return pipelines

    def create_s3_pipeline(
        self,
        pipeline_name: str,
        credentials: Optional[PipelineAwsCredential] = None,
        aws_credentials_source: Optional[str] = "user_provided",
        synthesize_files: Optional[bool] = False,
    ) -> S3Pipeline:
        """Create a new pipeline with files from Amazon S3.

        Parameters
        ----------
        pipeline_name: str
            The name of the pipeline.
        file_source: Optional[str]
            The type of location where the files to process are stored. Possible values are `local`, `aws`, `azure`, and `databricks`. The `local` option allows you to upload files from your local machine.
        credentials: PipelineAwsCredential
            The credentials to use to connect to AWS. Not required when `aws_credentials_source` is `from_environment`.
        synthesize_files: Optional[bool]
            Whether to generate a redacted version of the file in addition to the parsed output. Default value is `False`.
        aws_credentials_source: Optional[str]
           For an Amazon S3 pipeline, how to obtain the AWS credentials. Options are `user_provided` and `from_environment`. For `user_provided`, you provide the credentials in the `credentials` parameter. For `from_environment`, the credentials are read from your Textual instance.

        Returns
        -------
        S3Pipeline
            The newly created pipeline.
        """
        fs = FileSource.aws

        aws_cred_source = AwsCredentialsSource[aws_credentials_source]
        if (
            aws_cred_source == AwsCredentialsSource.user_provided
            and credentials is None
        ):
            raise BadArgumentsException(
                "Must specify AWS credentials when aws_credentials_source = 'user_provided'"
            )

        try:
            object_storage_type = ObjectStorageType.s3
            data = {
                "name": pipeline_name,
                "synthesizeFiles": synthesize_files,
                "fileSource": int(fs.value),
                "objectStorageType": int(object_storage_type.value),
            }

            if credentials is not None:
                data["parseJobExternalCredential"] = {
                    "credential": credentials,
                    "fileSource": int(fs.value),
                }
            if aws_credentials_source is not None and fs == FileSource.aws:
                data["awsCredentialSource"] = aws_cred_source

            p = self.client.http_post("/api/parsejobconfig", data=data)
            return S3Pipeline(p.get("name"), p.get("id"), self.client)
        except RequestException as req_err:
            if hasattr(req_err, "response") and req_err.response is not None:
                status_code = req_err.response.status_code
                error_message = req_err.response.text
                raise PipelineCreateError(f"Error {status_code}: {error_message}")
            else:
                raise req_err

    def create_azure_pipeline(
        self,
        pipeline_name: str,
        credentials: PipelineAzureCredential,
        synthesize_files: Optional[bool] = False,
    ) -> AzurePipeline:
        """Create a new pipeline with files from Azure blob storage.

        Parameters
        ----------
        pipeline_name: str
            The name of the pipeline.
        credentials: PipelineAzureCredential
            The credentials to use to connect to Azure.
        synthesize_files: Optional[bool]
            Whether to generate a redacted version of the file in addition to the parsed output. Default value is `False`.

        Returns
        -------
        AzurePipeline
            The newly created pipeline.
        """
        fs = FileSource.azure

        try:
            object_storage_type = ObjectStorageType.azure
            data = {
                "name": pipeline_name,
                "synthesizeFiles": synthesize_files,
                "fileSource": int(fs.value),
                "objectStorageType": int(object_storage_type.value),
                "parseJobExternalCredential": {
                    "credential": credentials,
                    "fileSource": int(fs.value),
                },
            }
            p = self.client.http_post("/api/parsejobconfig", data=data)
            return AzurePipeline(p.get("name"), p.get("id"), self.client)
        except RequestException as req_err:
            if hasattr(req_err, "response") and req_err.response is not None:
                status_code = req_err.response.status_code
                error_message = req_err.response.text
                raise PipelineCreateError(f"Error {status_code}: {error_message}")
            else:
                raise req_err

    def create_databricks_pipeline(
        self,
        pipeline_name: str,
        credentials: PipelineDatabricksCredential,
        synthesize_files: Optional[bool] = False,
    ) -> Pipeline:
        """Create a new pipeline on top of Databricks Unity Catalog.

        Parameters
        ----------
        pipeline_name: str
            The name of the pipeline.
        credentials: PipelineDatabricksCredential
            The credentials to use to connect to Databricks
        synthesize_files: Optional[bool]
            Whether to generate a redacted version of the file in addition to the parsed output. Default value is `False`.

        Returns
        -------
        Pipeline
            The newly created pipeline.
        """

        fs = FileSource.databricks

        try:
            object_storage_type = ObjectStorageType.databricks
            data = {
                "name": pipeline_name,
                "synthesizeFiles": synthesize_files,
                "fileSource": int(fs.value),
                "objectStorageType": int(object_storage_type.value),
                "parseJobExternalCredential": {
                    "credential": credentials,
                    "fileSource": int(fs.value),
                },
            }
            p = self.client.http_post("/api/parsejobconfig", data=data)
            return Pipeline(p.get("name"), p.get("id"), self.client)
        except RequestException as req_err:
            if hasattr(req_err, "response") and req_err.response is not None:
                status_code = req_err.response.status_code
                error_message = req_err.response.text
                raise PipelineCreateError(f"Error {status_code}: {error_message}")
            else:
                raise req_err

    def create_local_pipeline(
        self, pipeline_name: str, synthesize_files: Optional[bool] = False
    ) -> LocalPipeline:
        """Create a new pipeline from files uploaded from a local file system.

        Parameters
        ----------
        pipeline_name: str
            The name of the pipeline.
        synthesize_files: Optional[bool]
            Whether to generate a redacted version of the files in addition to the parsed output. Default value is `False`.

        Returns
        -------
        LocalPipeline
            The newly created pipeline.
        """

        try:
            p = self.client.http_post(
                "/api/parsejobconfig/local-files",
                data={"name": pipeline_name, "synthesizeFiles": synthesize_files},
            )
            return LocalPipeline(p.get("name"), p.get("id"), self.client)
        except RequestException as req_err:
            if hasattr(req_err, "response") and req_err.response is not None:
                status_code = req_err.response.status_code
                error_message = req_err.response.text
                raise PipelineCreateError(f"Error {status_code}: {error_message}")
            else:
                raise req_err

    def create_pipeline(self, pipeline_name: str):
        warn(
            "This method is deprecated. Instead, use the create_s3_pipeline, create_local_pipeline, create_azure_pipeline, and create_databricks_pipeline methods.",
            DeprecationWarning,
            stacklevel=1,
        )

    def delete_pipeline(self, pipeline_id: str):
        """Delete a pipeline.


        Parameters
        ----------
        pipeline_id: str
            The identifier of the pipeline.
        """

        try:
            result = self.client.http_delete(f"/api/parsejobconfig/{pipeline_id}")
            return result
        except RequestException as req_err:
            if hasattr(req_err, "response") and req_err.response is not None:
                status_code = req_err.response.status_code
                error_message = req_err.response.text
                raise PipelineDeleteError(f"Error {status_code}: {error_message}")
            else:
                raise req_err

    def get_pipeline_by_id(self, pipeline_id: str) -> Union[Pipeline, None]:
        """Gets the pipeline based on its identifier.

        Parameters
        ----------
        pipeline_id: str
            The identifier of the pipeline.

        Returns
        -------
        Union[Pipeline, None]
            The pipeline object, or None if no pipeline is found.
        """

        pipelines = self.get_pipelines()
        found_pipelines = list(filter(lambda x: x.id == pipeline_id, pipelines))
        if len(found_pipelines) == 0:
            return None

        if len(found_pipelines) > 1:
            raise Exception(
                "Found more than 1 pipeline with this identifier. This should not happen."
            )

        return found_pipelines[0]

    def parse_file(
        self, file: io.IOBase, file_name: str, timeout: Optional[int] = None
    ) -> FileParseResult:
        """Parse a given file. To open binary files, use the 'rb' option.

        Parameters
        ----------
        file: io.IOBase
            The opened file, available for reading, to parse.
        file_name: str
            The name of the file.
        timeout: Optional[int]
            Optional timeout in seconds. Indicates to stop waiting for the parsed result after the specified time.

        Returns
        -------
        FileParseResult
            The parsed document.
        """

        files = {
            "document": (
                None,
                json.dumps({"fileName": file_name, "csvConfig": {}}),
                "application/json",
            ),
            "file": file,
        }

        response = self.client.http_post(
            "/api/parse", files=files, timeout_seconds=timeout
        )
        document = response["document"]
        file_parse_result = response["fileParseResult"]

        return FileParseResult(
            file_parse_result, self.client, False, document=json.loads(document)
        )

    def parse_s3_file(
        self, bucket: str, key: str, timeout: Optional[int] = None
    ) -> FileParseResult:
        """Parse a given file found in Amazon S3. Uses boto3 to fetch files from Amazon S3.

        Parameters
        ----------
        bucket: str
            The bucket that contains the file to parse.
        key: str
            The key of the file to parse.
        timeout: Optional[int]
            Optional timeout in seconds. Indicates to stop waiting for parsed result after the specified time.

        Returns
        -------
        FileParseResult
            The parsed document.
        """

        import boto3

        s3 = boto3.resource("s3")
        obj = s3.Object(bucket, key)

        file_name = key.split("/")[-1]
        return self.parse_file(obj.get()["Body"].read(), file_name, timeout=timeout)


class TonicTextualParse(TextualParse):
    pass



from typing import List, Optional


class LabelCustomList:
    """
    Class to store the custom regular expression overrides (added or excluded values) to use during entity detection.

    Parameters
    ----------
    regexes : list[str]
        The list of regular expressions to use to override the original entity detection.

    """

    def __init__(self, regexes: Optional[List[str]] = None):
        self.regexes = regexes if regexes is not None else []

    def to_dict(self):
        return {"regexes": self.regexes}



from enum import Enum


class PdfTableCellTypeEnum(str, Enum):
    column_header = ("ColumnHeader",)
    row_header = ("RowHeader",)
    stub_head = ("StubHead",)
    description = ("Description",)
    content = ("Content",)



import re
from typing import List, Dict, Tuple, Union

MAX_HEADER_DEPTH = 6


def split_markdown(
    markdown: str, max_length: int
) -> List[Dict[str, Union[List[str], Tuple[int, int]]]]:
    return _split_by_header(markdown, max_length, 1, [], 0)


def _split_by_header(
    text: str,
    max_length: int,
    level: int,
    headers: List[str],
    base_index: int,
    max_header_depth: int = MAX_HEADER_DEPTH,
) -> List[Dict[str, Union[List[str], Tuple[int, int]]]]:
    header_pattern = re.compile(r"^(#{1," + str(level) + r"})\s+(.*)$", re.MULTILINE)
    sections = []
    last_index = 0

    for match in header_pattern.finditer(text):
        if match.start() > last_index:
            sections.append((last_index, match.start(), headers))
        headers = headers[: level - 1] + [match.group(2)]
        last_index = match.start()

    if last_index < len(text):
        sections.append((last_index, len(text), headers))

    split_sections = []
    for start, end, sec_headers in sections:
        section_length = end - start
        if section_length > max_length:
            if level < max_header_depth:
                split_sections.extend(
                    _split_by_header(
                        text[start:end],
                        max_length,
                        level + 1,
                        sec_headers,
                        base_index + start,
                    )
                )
            else:
                split_sections.extend(
                    _split_large_section(
                        text[start:end], max_length, sec_headers, base_index + start
                    )
                )
        else:
            split_sections.append(
                {
                    "headers": sec_headers,
                    "indices": (base_index + start, base_index + end),
                }
            )
    return split_sections


def _split_large_section(
    text: str, max_length: int, headers: List[str], base_index: int
) -> List[Dict[str, Union[List[str], Tuple[int, int]]]]:
    chunks = []
    start = 0

    while start < len(text):
        end = start + max_length
        if end >= len(text):
            chunks.append(
                {
                    "headers": headers,
                    "indices": (base_index + start, base_index + len(text)),
                }
            )
            break

        newline_pos = text.rfind("\n", start, end)
        if newline_pos != -1 and newline_pos > start:
            chunks.append(
                {
                    "headers": headers,
                    "indices": (base_index + start, base_index + newline_pos),
                }
            )
            start = newline_pos + 1
        else:
            chunks.append(
                {"headers": headers, "indices": (base_index + start, base_index + end)}
            )
            start = end

    return chunks



from typing import Dict
from requests.exceptions import HTTPError, RequestException


class DatasetNameAlreadyExists(Exception):
    """
    Raised when there is an attempt to create a dataset with a name that already exists.
    """

    def __init__(self, errors):
        # Call the base class constructor with the parameters it needs
        super().__init__(
            "The dataset name already exists. Dataset names must be unique. Choose a "
            "different name."
        )
        self.errors = errors


class DatasetFileMatchesExistingFile(HTTPError):
    """
    Raised when the content in a file to upload matches the content in an existing file
    in the dataset.
    """

    def __init__(self, errors):
        message = (
            "The file content matches content in an existing dataset file. Choose a "
            "different file."
        )

        super().__init__(errors.response.content or message)
        self.errors = errors


class InvalidJsonForRedactionRequest(Exception):
    """
    Raised when the JSON redaction request contains invalid JSON
    """

    def __init__(self, msg):
        super().__init__(msg)


class LicenseInvalid(HTTPError):
    """
    Raised when either your license has expired OR you've exceeded your allowed word
    limit
    """

    def __init__(self, errors):
        super().__init__(
            str(errors.response.content)
            or "Invalid Textual license. Please reach out to textual@tonic.ai."
        )
        self.errors = errors


class FileNotReadyForDownload(Exception):
    """
    Raised when you make a request to download a file that is not yet ready for download
    """

    def __init__(self, msg):
        super().__init__(msg)


class BadArgumentsException(Exception):
    def __init__(self, msg):
        super().__init__(msg)


class ErrorWhenDownloadFile(HTTPError):
    """
    Raised when server returns 500 when attempting to download file
    """

    def __init__(self, errors):
        super().__init__(
            "Either error occurred while downloading file or the file redaction job was "
            "cancelled."
        )
        self.errors = errors


class BadRequestDownloadFile(HTTPError):
    """
    Raised when server returns 400 when attempting to download file
    """

    def __init__(self, msg, response=None):
        super().__init__(msg)
        self.response = response


class JobsNotSuccessful(HTTPError):
    """
    Raised when at least one of the jobs in the comparison has not completed successfully
    """

    def __init__(self, errors):
        super().__init__("Both jobs must have completed successfully")
        self.response = errors.response
        self.request = errors.request


class FileUploadError(RequestException):
    """
    Raised when server returns an error when uploading a file
    """

    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class DatasetFileNotFound(RequestException):
    """
    Raised when an action is taken on a dataset file that cannot be found in the provided Dataset
    """

    def __init__(self, dataset_name: str, file_id: str):
        message = f"Dataset {dataset_name} does not have a file with id {file_id}"
        super().__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class RunPipelineError(RequestException):
    """
    Raised when server returns an error when running pipeline
    """

    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class PipelineDeleteError(RequestException):
    """
    Raised when server returns 500 when deleting pipeline
    """

    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class PipelineCreateError(RequestException):
    """
    Raised when server returns 500 when creating pipeline
    """

    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class DownloadResultFileError(RequestException):
    """
    Raised when server returns 500 when downloading file
    """

    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __str__(self):
        return self.message


class ParseFileTimeoutException(Exception):
    """
    Raised when parsing an uploaded file takes too long.
    """

    def __init__(self):
        super().__init__(
            "Parsing file took too long.  Either raise the timeout limit by modifying the TONIC_TEXTUAL_PARSE_TIMEOUT_IN_SECONDS or process your file via our pipeline which has no time limits."
        )


class TextualServerError(Exception):
    """
    Raised when the Textual server responds with a 500.
    """

    def __init__(self, error_payload: Dict):
        msg = error_payload["error"]
        super().__init__(msg)



from typing import List, Optional
from tonic_textual.classes.file_content.base_document import BaseDocument
from tonic_textual.classes.file_content.content import Content
from tonic_textual.classes.enums.pdf_content_type import PdfContentTypeEnum
from tonic_textual.classes.enums.pdf_table_cell_type import PdfTableCellTypeEnum
from tonic_textual.classes.httpclient import HttpClient
from tonic_textual.classes.table import Table


class PdfPageContent:
    def __init__(self, client: HttpClient, json_def):
        self.type: PdfContentTypeEnum = json_def["type"]
        self.content: Content = Content(client, json_def["content"])


class PdfTableCell:
    def __init__(self, client: HttpClient, json_def):
        self.type: PdfTableCellTypeEnum = json_def["type"]
        self.content: Content = Content(client, json_def["content"])


class PdfTable:
    def __init__(self, client: HttpClient, json_def):
        self.rows: List[List[Optional[PdfTableCell]]] = [
            [PdfTableCell(client, f) if f is not None else None for f in row]
            for row in json_def
        ]


class PdfDocument(BaseDocument):
    def __init__(self, client: HttpClient, json_def):
        super().__init__(client, json_def)
        self.document: Content = Content(client, json_def["content"])
        self.pages: List[List[PdfPageContent]] = [
            [PdfPageContent(client, c) for c in p] for p in json_def["pages"]
        ]

        self.tables: List[Table] = []
        for idx, table in enumerate(json_def["tables"]):
            header = []
            first_row = table[0] if len(table) > 0 else []
            for col in first_row:
                if col is None or col["type"] == "ColumnHeader":
                    header.append(
                        col["content"]["text"]
                        if col is not None and col["type"] == "ColumnHeader"
                        else ""
                    )

            self.tables.append(
                Table(
                    [
                        [
                            Content(client, f["content"] if f is not None else None)
                            for f in row
                        ]
                        for row in table
                    ],
                    f"table_{idx}",
                    header,
                )
            )



from enum import Enum

class docx_image_policy(Enum):
    redact = 1
    ignore = 2
    remove = 3

class docx_comment_policy(Enum):
    remove = 1
    ignore = 2

class pdf_signature_policy(Enum):
    redact = 1
    ignore = 2



from typing import Optional


class PipelineAwsCredential(dict):
    """Class to represent AWS credentials used to create an Amazon S3 pipeline.

    Parameters
    ----------
    aws_access_key_id: str
        The AWS access key ID. The access key ID should have permissions on the S3 buckets that you plan to read data from and write data to.

    aws_secret_access_key: str
        The AWS secret access key. The secret access key should have permissions on the S3 buckets that you plan to read data from and write data to.

    aws_region: str
        The AWS region

    aws_session_token: Optional[str]
        An optional AWS session token. Only needed if required by your IAM credentials.
    """

    def __init__(
        self,
        aws_access_key_id: str,
        aws_secret_access_key: str,
        aws_region: str,
        aws_session_token: Optional[str] = None,
    ):
        self.aws_access_key = aws_access_key_id
        self.aws_secret_key = aws_secret_access_key
        self.aws_region = aws_region
        self.aws_session_token = aws_session_token

        dict.__init__(
            self,
            accessKey=aws_access_key_id,
            secretKey=aws_secret_access_key,
            region=aws_region,
            sessionToken=aws_session_token,
        )

    def to_dict(self):
        return {
            "accessKey": self.aws_access_key,
            "secretKey": self.aws_secret_key,
            "region": self.aws_region,
            "sessionToken": self.aws_session_token,
        }



from typing import Dict, List
from tonic_textual.classes.common_api_responses.single_detection_result import (
    SingleDetectionResult,
)
from tonic_textual.classes.httpclient import HttpClient


class BaseDocument:
    def __init__(self, client: HttpClient, json_def: Dict):
        self.client = client
        self._json_def = json_def
        self.markdown = json_def["content"]["text"]
        self.entities: List[SingleDetectionResult] = [
            SingleDetectionResult(
                s["start"], s["end"], s["label"], s["text"], s["score"]
            )
            for s in json_def["content"]["entities"]
        ]
        self.schema_version = json_def.get("schemaVersion", 0)

    def get_markdown(self):
        return self.markdown

    def get_all_entities(self):
        return self.entities

    def get_json(self) -> Dict:
        return self._json_def

    def to_dict(self):
        return self._json_def



from typing import List

import requests

from tonic_textual.classes.httpclient import HttpClient
from tonic_textual.classes.parse_api_responses.file_parse_result import FileParseResult
from tonic_textual.classes.parse_api_responses.file_parse_results_diff import (
    FileParseResultsDiff,
)
from tonic_textual.classes.tonic_exception import JobsNotSuccessful


class FileParseResultsDiffEnumerator(object):
    """Enumerates the files in a diff between two jobs.

    Parameters
    ----------
    job_id1: str
        The first job identifier.

    job_id2: str
        The second job identifier.

    client: HttpClient
        The HTTP client to use.
    """

    def __init__(self, job_id1: str, job_id2: str, client: HttpClient):
        self.job_id1 = job_id1
        self.job_id2 = job_id2
        self.client = client

    def __iter__(self):
        self.pagination_token: str = "0"
        self.available_files: List[FileParseResultsDiff] = []
        self.idx = 0
        return self

    def __next__(self):
        return self.next()

    def next(self) -> FileParseResultsDiff:
        if self.idx < len(self.available_files):
            file_to_return = self.available_files[self.idx]
            self.idx = self.idx + 1
            return file_to_return

        if self.pagination_token is None:
            raise StopIteration()

        with requests.Session() as session:
            url = "/api/parsejob/"
            url += f"{self.job_id1}/diff/{self.job_id2}?offset={self.pagination_token}"
            try:
                response = self.client.http_get(
                    url,
                    session=session,
                )
            except requests.exceptions.HTTPError as e:
                if "Both jobs must have completed successfully" in e.response.text:
                    raise JobsNotSuccessful(e)
            self.pagination_token = str(response["offset"] + len(response["records"]))
            files = response["records"]
            self.available_files = []
            for file in files:
                file_parse_result = FileParseResult(file["file"], self.client)
                self.available_files.append(
                    FileParseResultsDiff(file["status"], file_parse_result)
                )

        if len(self.available_files) == 0:
            raise StopIteration()

        self.idx = 1
        return self.available_files[0]



from typing import Dict


class BaseFile(object):
    def __init__(self, response: Dict):
        self.id = response["id"]
        self.userId = response["userId"]
        self.oid = response["oid"]
        self.fileSizeInKb = response["fileSizeInKb"]
        self.fileName = response["fileName"]
        self.escapeChar = response["escapeChar"]
        self.quoteChar = response["quoteChar"]
        self.hasHeader = response["hasHeader"]
        self.delimiter = response["delimiter"]
        self.nullChar = response["nullChar"]
        self.numRows = response["numRows"]
        self.fileType = response["fileType"]
        self.columnCount = response["columnCount"]
        self.piiTypeCounts = response["piiTypeCounts"]
        self.wordCount = response["wordCount"]
        self.characterCount = response["characterCount"]
        self.redactedWordCount = response["redactedWordCount"]
        self.piiTypes = response["piiTypes"]
        self.piiTypeExamples = response["piiTypeExamples"]
        self.createdDate = response["createdDate"]
        self.lastModifiedDate = response["lastModifiedDate"]
        self.fileHash = response["fileHash"]
        self.fileSource = response["fileSource"]
        self.filePath = response["filePath"]

    def to_dict(self):
        return {
            "textual_id": self.id,
            "userId": self.userId,
            "fileName": self.fileName,
            "fileType": self.fileType,
            "wordCount": self.wordCount,
            "characterCount": self.characterCount,
            "redactedWordCount": self.redactedWordCount,
            "createdDate": self.createdDate,
            "lastModifiedDate": self.lastModifiedDate,
            "fileHash": self.fileHash,
            "fileSource": self.fileSource,
            "filePath": self.filePath,
        }



from typing import List

from tonic_textual.classes.common_api_responses.replacement import Replacement


class BulkRedactionResponse(dict):
    """Bulk Redaction response object

    Attributes
    ----------
    bulk_text : List[str]
        The original text.
    bulk_redacted_text : List[str]
        The redacted and synthesized text.
    usage : int
        The number of words used
    de_identify_results : List[Replacement]
        The list of named entities that were found in bulk_text.
    """

    def __init__(
        self,
        bulk_text: List[str],
        bulk_redacted_text: List[str],
        usage: int,
        de_identify_results: List[Replacement],
    ):
        self.bulk_text = bulk_text
        self.bulk_redacted_text = bulk_redacted_text
        self.usage = usage
        self.de_identify_results = de_identify_results
        dict.__init__(
            self,
            bulk_text=bulk_text,
            bulk_redacted_text=bulk_redacted_text,
            usage=usage,
            de_identify_results=de_identify_results,
        )

    def describe(self) -> str:

        result = ""
        for redacted_text, de_id_res in zip(
            self.bulk_redacted_text, self.de_identify_results
        ):
            result += f"{redacted_text}\n"
            for replacement in de_id_res:
                result += f"{replacement.describe()}\n"
        return result

    def get_usage(self):
        return self.usage



from datetime import datetime
from typing import Dict, List, Optional, Union

import requests
from requests import RequestException

from tonic_textual.classes.common_api_responses.base_file import BaseFile
from tonic_textual.classes.common_api_responses.single_detection_result import (
    SingleDetectionResult,
)
from tonic_textual.classes.enums.file_type import FileTypeEnum
from tonic_textual.classes.file_content.csv_document import CsvDocument
from tonic_textual.classes.file_content.docx_document import DocxDocument
from tonic_textual.classes.file_content.pdf_document import PdfDocument
from tonic_textual.classes.file_content.raw_document import RawDocument
from tonic_textual.classes.file_content.xlsx_document import XlsxDocument
from tonic_textual.classes.httpclient import HttpClient
from tonic_textual.classes.table import Table
from tonic_textual.classes.tonic_exception import DownloadResultFileError
from tonic_textual.enums.pii_state import PiiState
from tonic_textual.markdown_utils import split_markdown
from tonic_textual.generator_utils import (
    filter_entities_by_config,
    make_utf_compatible_entities,
)


class FileParseResult(object):
    """A class that represents the result of a parsed file.

    Parameters
    ----------
    response: Dict
        The response from the API.

    client: HttpClient
        The HTTP client to use.

    lazy_load_content: bool
        Whether to lazy load the content of the file. Default is False.
    """

    def __init__(
        self,
        response: Dict,
        client: HttpClient,
        lazy_load_content=False,
        document: Dict = None,
    ):
        self.id: str = response["id"]
        self.file: BaseFile = BaseFile(response["file"])
        self.file_parse_job_id: str = response["fileParseJobId"]
        self.parsed_file_path: str = response["parsedFilePath"]
        self.created_date: datetime = response["createdDate"]
        self.last_modified_date: datetime = response["lastModifiedDate"]
        self.__is_content_retrieved = False
        self.client = client

        if document is not None:
            self.content = self.__convert_document_json(document)
            self.__is_content_retrieved = True

        if not lazy_load_content and self.__is_content_retrieved is False:
            self.__get_content()
        self.metadata = {}

    def __convert_document_json(
        self, doc_json: Dict
    ) -> Union[RawDocument, CsvDocument, DocxDocument, XlsxDocument, PdfDocument]:
        if self.file.fileType == FileTypeEnum.csv:
            return CsvDocument(self.client, doc_json)
        elif self.file.fileType == FileTypeEnum.raw:
            return RawDocument(self.client, doc_json)
        elif self.file.fileType == FileTypeEnum.xlsx:
            return XlsxDocument(self.client, doc_json)
        elif self.file.fileType == FileTypeEnum.docX:
            return DocxDocument(self.client, doc_json)
        elif (
            self.file.fileType == FileTypeEnum.pdf
            or self.file.fileType == FileTypeEnum.png
            or self.file.fileType == FileTypeEnum.jpg
            or self.file.fileType == FileTypeEnum.tif
        ):
            return PdfDocument(self.client, doc_json)
        else:
            raise Exception("Unknown file type " + self.file.fileType)

    def __get_content(self):
        if not self.__is_content_retrieved:
            with requests.Session() as session:
                response = self.client.http_get(
                    f"/api/parsejob/files/{self.id}/content", session=session
                )

                self.content = self.__convert_document_json(response)
                self.__is_content_retrieved = True

    def get_json(self) -> Dict:
        """Returns the raw JSON generated by Tonic Textual.

        Returns
        -------
        Dict
            The raw JSON that Textal generates when it parses the file, in the form of a dictionary.
        """

        self.__get_content()
        return self.content.get_json()

    def get_markdown(
        self,
        generator_config: Dict[str, PiiState] = dict(),
        generator_default: PiiState = PiiState.Off,
        random_seed: Optional[int] = None
    ) -> str:
        """Returns the file in Markdown format. In the file, the entities are redacted or synthesized based on the specified configuration.

        Parameters
        ----------
        generator_config: Dict[str, PiiState]
            A dictionary of sensitive data entity types. For each entity type, indicates whether
            to redact, synthesize, or ignore the detected entities.
            Values must be one of "Redaction", "Synthesis", or "Off".

        generator_default: PiiState = PiiState.Redaction
            The default redaction to use for all entity types that not specified in generator_config.
            Value must be one of "Redaction", "Synthesis", or "Off".
        
        random_seed: Optional[int] = None
            An optional value to use to override Textual's default random number
            seeding. Can be used to ensure that different API calls use the same or
            different random seeds.
            
        Returns
        -------
        str
            The file in Markdown format. In the file, the entities are redacted or synthesized based on
            generator_config and generator_default.
        """
        self.__get_content()
        rawtext = self.content.get_markdown()
        filtered_entities = filter_entities_by_config(
            self.content.get_all_entities(),
            generator_config,
            generator_default,
        )
        if len(filtered_entities) == 0:
            return rawtext
        utf_compatible_filtered_entities = make_utf_compatible_entities(
            rawtext, filtered_entities
        )

        if random_seed is not None:
            additional_headers = {"textual-random-seed": str(random_seed)}
        else:
            additional_headers = {}

        response = self.client.http_post(
            "/api/redact/known_entities",
            data={
                "knownEntities": utf_compatible_filtered_entities,
                "text": rawtext,
                "generatorConfig": generator_config,
                "generatorDefault": generator_default,
            },
            additional_headers=additional_headers
        )
        return response["redactedText"]

    def describe(self) -> str:
        """Returns the parsed file path."""
        return f"{self.parsed_file_path}"

    def get_all_entities(self) -> List[SingleDetectionResult]:
        """Returns a list of all of the detected entities in the file.

        Returns
        -------
        List[SingleDetectionResult]
            A list of detected entities in the file.
        """
        self.__get_content()
        return self.content.get_all_entities()

    def get_entities(
        self,
        generator_config: Dict[str, PiiState] = dict(),
        generator_default: PiiState = PiiState.Redaction,
        allow_overlap: bool = False,
    ) -> List[SingleDetectionResult]:
        """Returns a list of entities in the document. The entities are filtered by the
        generator_default configuration.

        Parameters
        ----------
        generator_default: PiiState
            The default redaction to use for all entity types that not specified in generator_config.
            Value must be one of "Redaction", "Synthesis", or "Off".

        generator_config: Dict[str, PiiState]
            A dictionary of sensitive data entity types. For each entity type, indicates whether
            to redact, synthesize, or ignore the detected entities.
            Values must be one of "Redaction", "Synthesis", or "Off".

        Returns
        -------
        List[SingleDetectionResult]
            A list of the detected entities. Each item in list contains the entity type,
            source start index, source end index, the entity text, and replacement text.
        """
        self.__get_content()
        all_entities = self.content.get_all_entities()
        rawtext = self.content.get_markdown()
        filtered_entities = filter_entities_by_config(
            all_entities, generator_config, generator_default
        )
        if allow_overlap:
            return filtered_entities
        utf_compatible_filtered_entities = make_utf_compatible_entities(
            rawtext, filtered_entities
        )
        response = self.client.http_post(
            "/api/redact/known_entities",
            data={"knownEntities": utf_compatible_filtered_entities},
        )
        return [
            SingleDetectionResult(
                x["start"], x["end"], x["label"], x["text"], x["score"]
            )
            for x in list(response["deIdentifyResults"])
        ]

    def is_sensitive(
        self, sensitive_entity_types: List[str], start: int = 0, end: int = -1
    ) -> bool:
        """Returns True if the element contains sensitive data. Otherwise returns False.

        Parameters
        ----------
        sensitive_entity_types: List[str]
            A list of sensitive entity types to check for.

        start: int = 0
            The start index to check for sensitive data.

        end: int = -1
            The end index to check for sensitive data.

        Returns
        -------
        bool
            Returns True if the element contains sensitive data. Otherwise returns False.
        """
        all_entities = self.get_entities()
        entities = [
            ent for ent in all_entities if ent["label"] in sensitive_entity_types
        ]
        filtered_entities = self._find_intersecting_entites(start, end, entities)
        return len(filtered_entities) > 0

    def get_tables(self) -> List[Table]:
        """Returns a list of tables found in the document. Applies to CSV, XLSX, PDF, and image files.

        Parameters
        ----------
        sensitive_entity_types: List[str]
            A list of sensitive entity types to check for.

        start: int = 0
            The start index to check for sensitive data.

        end: int = -1
            The end index to check for sensitive data.

        Returns
        -------
        bool
            Returns True if the element contains sensitive data. Otherwise returns False.
        """
        self.__get_content()

        if self.file.fileType not in (
            FileTypeEnum.csv,
            FileTypeEnum.xlsx,
            FileTypeEnum.pdf,
            FileTypeEnum.png,
            FileTypeEnum.tif,
            FileTypeEnum.jpg,
        ):
            return []

        return self.content.tables

    @staticmethod
    def _make_entity_metadata(entity_list: List[Dict]) -> Dict[str, List[str]]:
        """Flattens a list of entity spans into a dictionary of entities per type"""
        entity_dict = {}
        for entity in entity_list:
            entity_type = entity["label"]
            entity_list = entity_dict.get(entity_type, [])
            entity_text = entity["text"]
            if entity_text not in entity_list:
                entity_list.append(entity_text)
            entity_dict[entity_type] = entity_list
        return entity_dict

    @staticmethod
    def _find_intersecting_entites(
        start: int, end: int, entities: List[SingleDetectionResult]
    ) -> List[Dict]:
        """Finds entities that intersect with the given start and end gindices
        Parameters
        ----------
        start: int
            The start index to check for intersecting entities.
        end: int
            The end index to check for intersecting entities. If -1, returns all
            entities that occur after start.
        Returns
        -------
        List[Dict]
            A list of entities that intersect with the given start and end indices.
        """
        intersecting_entities = []
        for entity in entities:
            entity_start = entity["start"]
            entity_end = entity["end"]
            if (entity_start <= end or end == -1) and entity_end >= start:
                intersecting_entities.append(entity)
        return intersecting_entities

    def get_chunks(
        self,
        max_chars=15_000,
        generator_config: Dict[str, PiiState] = {},
        generator_default: PiiState = PiiState.Off,
        metadata_entities: List[str] = [],
        include_metadata=True,
    ) -> List:
        """Returns a list of chunks of text from the document. The chunks are filtered
        by the generator_default configuration.

        Parameters
        ----------
        max_chars: int = 15_000
            The maximum number of characters in each chunk.

        generator_config: Dict[str, PiiState]
            A dictionary of sensitive data entity types. For each entity type, indicates
            whether to redact, synthesize, or ignore the detected entities.
            Values must be one of "Redaction", "Synthesis", or "Off".

        generator_default: PiiState = PiiState.Redaction
            The default redaction to use for all entity types that are not specified in
            generator_config.
            Value must be one of "Redaction", "Synthesis", or "Off".


        include_metadata: bool = True
            If True, the metadata is included in the chunk.
        Returns:
            List[str]
                A list of strings that contain the chunks of text.
        """
        text = self.get_markdown(generator_config, generator_default)
        all_entities = self.get_all_entities()
        entities = [ent for ent in all_entities if ent["label"] in metadata_entities]
        output = []
        for chunk in split_markdown(text, max_chars):
            start, end = chunk["indices"]
            headers = chunk["headers"]
            sensitive_entity_types = [
                label for label, key in generator_config.items() if key != PiiState.Off
            ]

            chunk_text = text[start:end]
            is_sensitive = self.is_sensitive(
                sensitive_entity_types,
                start,
                end,  # type: ignore
            )
            chunk_dict = {"text": chunk_text, "is_sensitive": is_sensitive}
            if include_metadata:
                entity_metdata = self._make_entity_metadata(  # type: ignore
                    self._find_intersecting_entites(
                        start,
                        end,
                        entities,  # type: ignore
                    )
                )

                metadata = {"headers": headers, "entities": entity_metdata}
                metadata.update(self.metadata)
                chunk_dict["metadata"] = metadata
            output.append(chunk_dict)
        return output

    def download_results(self) -> str:
        """Downloads the results file.

        Returns
        -------
        string
            The results file.
        """
        try:
            with requests.Session() as session:
                return self.client.http_get_file(
                    f"/api/parsejob/files/{self.id}/content", session=session
                ).decode("utf-8")
        except RequestException as req_err:
            if hasattr(req_err, "response") and req_err.response is not None:
                status_code = req_err.response.status_code
                error_message = req_err.response.text
                raise DownloadResultFileError(f"Error {status_code}: {error_message}")
            else:
                raise req_err



from typing import Dict, List

from tonic_textual.classes.common_api_responses.single_detection_result import (
    SingleDetectionResult,
)
from tonic_textual.enums.pii_state import PiiState


def utf16len(c):
    """Returns the length of the single character 'c'
    in UTF-16 code units."""
    return 1 if ord(c) < 65536 else 2


def filter_entities_by_config(
    entities: List[SingleDetectionResult],
    generator_config: Dict[str, PiiState],
    generator_default: PiiState,
) -> List[SingleDetectionResult]:
    filtered_entities = []
    for entity in entities:
        if entity["label"] in generator_config:
            if generator_config[entity["label"]] == PiiState.Off:
                continue
        elif generator_default == PiiState.Off:
            continue
        filtered_entities.append(entity)
    return filtered_entities


def make_utf_compatible_entities(
    text: str, entities: List[SingleDetectionResult]
) -> List[Dict]:
    offsets = []
    prev = 0
    for c in text:
        offset = utf16len(c) - 1
        offsets.append(prev + offset)
        prev = prev + offset

    utf_compatible_entities = []
    for entity in entities:
        new_entity = entity.to_dict()
        new_entity["pythonStart"] = entity["start"]
        new_entity["pythonEnd"] = entity["end"]
        new_entity["start"] = entity["start"] + offsets[entity["start"]]
        new_entity["end"] = entity["end"] + offsets[entity["end"] - 1]
        utf_compatible_entities.append(new_entity)

    return utf_compatible_entities


def validate_generator_options(
    generator_default: PiiState, generator_config: Dict[str, PiiState]
) -> None:
    invalid_pii_states = [
        v for v in list(generator_config.values()) if v not in PiiState._member_names_
    ]
    if len(invalid_pii_states) > 0:
        raise Exception(
            "Invalid configuration for generator_config. The allowed values are "
            "Off, Synthesis, and Redaction."
        )
    if generator_default not in PiiState._member_names_:
        raise Exception(
            "Invalid option for generator_default. The allowed values are Off, "
            "Synthesis, and Redaction."
        )



from typing import List, Optional
from warnings import warn

import requests
import io

from requests import RequestException

from tonic_textual.classes.SolarCsvConfig import SolarCsvConfig
from tonic_textual.classes.file_parse_result_diff_enumerator import (
    FileParseResultsDiffEnumerator,
)
from tonic_textual.classes.httpclient import HttpClient
from tonic_textual.classes.pipeline_file_enumerator import PipelineFileEnumerator
from tonic_textual.classes.pipeline_run import PipelineRun
from tonic_textual.classes.tonic_exception import FileUploadError
from abc import ABC


class Pipeline(ABC):
    """Class to represent and provide access to a Tonic Textual pipeline. This class is abstract. Do not instantiate it directly.

    Parameters
    ----------
    name: str
        Pipeline name.

    id: str
        Pipeline identifier.

    client: HttpClient
        The HTTP client to use.
    """

    def __init__(self, name: str, id: str, client: HttpClient):
        self.id = id
        self.name = name
        self.client = client

    def describe(self) -> str:
        """Returns the name and id of the pipeline."""
        description = "--------------------------------------------------------\n"
        description += f"Name: {self.name}\n"
        description += f"ID: {self.id}\n"
        description += "--------------------------------------------------------\n"
        return description

    def get_runs(self) -> List[PipelineRun]:
        """Get the runs for the pipeline.

        Returns
        -------
        List[PipelineRun]
            A list of PipelineRun objects.
        """
        with requests.Session() as session:
            response = self.client.http_get(
                f"/api/parsejob/{self.id}/jobs", session=session
            )
            runs: List[PipelineRun] = []
            for run in response:
                runs.append(
                    PipelineRun(run["id"], run["status"], run["endTime"], self.client)
                )
            return runs

    def run(self) -> str:
        """Run the pipeline.

        Returns
        -------
        str
            The ID of the job.
        """
        try:
            response = self.client.http_post(f"/api/parsejobconfig/{self.id}/start")
            return response
        except RequestException as req_err:
            if hasattr(req_err, "response") and req_err.response is not None:
                status_code = req_err.response.status_code
                error_message = req_err.response.text
                raise FileUploadError(f"Error {status_code}: {error_message}")
            else:
                raise req_err

    def enumerate_files(self, lazy_load_content=True) -> PipelineFileEnumerator:
        """Enumerate the files in the pipeline.

        Parameters
        ----------
        lazy_load_content: bool
            Whether to lazily load the content of the files. Default is True.

        Returns
        -------
        PipelineFileEnumerator
            An enumerator for the files in the pipeline.
        """
        runs = self.get_runs()
        successful_runs = filter(lambda r: r.status == "Completed", runs)
        sorted_finished_runs = sorted(
            successful_runs, key=lambda r: r.end_time, reverse=True
        )

        if len(sorted_finished_runs) == 0:
            return PipelineFileEnumerator(
                "", self.client, lazy_load_content=lazy_load_content
            )

        job_id = sorted_finished_runs[0].id
        return PipelineFileEnumerator(
            job_id, self.client, lazy_load_content=lazy_load_content
        )

    def get_delta(
        self, pipeline_run1: PipelineRun, pipeline_run2: PipelineRun
    ) -> FileParseResultsDiffEnumerator:
        """Enumerates the files in the diff between two pipeline runs.

        Parameters
        ----------
        pipeline_run1: PipelineRun
            The first pipeline run.

        pipeline_run2: PipelineRun
            The second pipeline run.

        Returns
        -------
        FileParseResultsDiffEnumerator
            An enumerator for the files in the diff between the two runs.
        """
        return FileParseResultsDiffEnumerator(
            pipeline_run1.id, pipeline_run2.id, self.client
        )

    def upload_file(
        self,
        file: io.IOBase,
        file_name: str,
        csv_config: Optional[SolarCsvConfig] = None,
    ) -> str:
        warn(
            "This method has been deprecated. Please use the new add_file method",
            DeprecationWarning,
            stacklevel=1,
        )

    def set_synthesize_files(self, synthesize_files: bool):
        self.client.http_patch(
            f"/api/parsejobconfig/{self.id}", data={"synthesizeFiles": synthesize_files}
        )



from tonic_textual.classes.enums.file_source import FileSource


class PipelineAzureCredential(dict):
    """Class to represent Azure credentials used to create an Azure pipeline.

    Parameters
    ----------
    account_name: str
        The name of the Azure account from where you plan to access files in Azure Blob Storage.

    account_key: str
        The account key. The key must have access to the Azure Blob Storage containers that you plan to read data from and write data to.
    """

    def __init__(self, account_name: str, account_key: str):
        self.file_source = FileSource.azure

        self.account_name = account_name
        self.account_key = account_key

        dict.__init__(self, accountName=account_name, accountKey=account_key)

    def to_dict(self):
        return {"accountName": self.account_name, "accountKey": self.account_key}



from typing import Dict, List

from tonic_textual.classes.file_content.content import Content


class Table(dict):
    def __init__(self, data: List[List[Content]], table_name: str, header: List[str]):
        self.table_name = table_name
        self.data = data
        self.header = header
        dict.__init__(self, table_name=table_name, data=data, header=header)

    def get_data(self) -> List[List[str]]:
        return [[col.text for col in row] for row in self.data]

    def to_dict(self) -> Dict:
        return {
            "table_name": self.table_name,
            "data": self.data,
            "header": self.header,
        }
